import collections
import collections.abc
import datetime
import itertools as it
import os.path
import re

from PySide6 import QtCore, QtWidgets

from ...config.configuration import config
from ...config.cutoff_history import HistoryEntry, cutoff_history
from ...config.cutoff_presets import Preset, cutoff_presets
from ...core import calculation, file
from ..cutoff_preset_combobox import CutoffPresetComboBox
from ..cutoff_table_widget import CutoffTableWidget
from .cutoff_history_dialog import CutoffHistoryDialog
from .util.calc_table import CalculationTable, TableModel


class CalculationSettingsDialog(QtWidgets.QDialog):

    RES_MIN = 32
    RES_MAX = 1024
    RES_INTERVAL = 32

    def __init__(self, parent, file_frame_dict):
        super().__init__(parent)

        self.control = parent.control
        self.resolution = config.Computation.std_resolution
        self.filenames = [f for f in file_frame_dict.keys()]
        self.file_frame_dict = file_frame_dict

        self.init_gui()
        self.setWindowTitle("Calculation Settings")

    def init_gui(self):

        vbox = QtWidgets.QVBoxLayout()
        hbox = QtWidgets.QHBoxLayout()
        inner_layout = QtWidgets.QVBoxLayout()
        button_hbox = QtWidgets.QHBoxLayout()
        res_hbox = QtWidgets.QHBoxLayout()

        self.res_slider = QtWidgets.QSlider(QtCore.Qt.Horizontal, self)
        self.res_slider.setMinimum(0)
        self.res_slider.setMaximum((self.RES_MAX - self.RES_MIN) / self.RES_INTERVAL)
        # self.res_slider.setTickInterval(1)
        # self.res_slider.setTickPosition(QtWidgets.QSlider.TicksBelow)
        self.res_slider.valueChanged[int].connect(self.slider_changing)
        self.res_slider.sliderReleased.connect(self.slider_released)

        self.lineedit = QtWidgets.QLineEdit(self)
        self.lineedit.setFocusPolicy(QtCore.Qt.StrongFocus)
        self.lineedit.setMinimumSize(30, 1)
        self.lineedit.setMaximumSize(40, 40)
        self.lineedit.returnPressed.connect(self.lineedit_return)

        res_hbox.addWidget(QtWidgets.QLabel("resolution:", self))
        res_hbox.addWidget(self.res_slider)
        res_hbox.addWidget(self.lineedit)

        # set font
        #        font = QFont("Courier New", 14)
        #        table_view.setFont(font)
        # set column width to fit contents (set font first!)
        # enable sortingq
        #        table_view.setSortingEnabled(True)

        ok_button = QtWidgets.QPushButton("Ok", self)
        ok_button.setAutoDefault(False)
        ok_button.clicked.connect(self.ok)

        cancel_button = QtWidgets.QPushButton("Cancel", self)
        cancel_button.setAutoDefault(False)
        cancel_button.clicked.connect(self.cancel)

        button_hbox.addStretch()
        button_hbox.addWidget(ok_button)
        button_hbox.addStretch()
        button_hbox.addWidget(cancel_button)
        button_hbox.addStretch()

        self.table_view = CalculationTable(self)
        self.res_slider.setValue((self.resolution - self.RES_MIN) / self.RES_INTERVAL)
        self.update_table()

        def create_separator():
            line = QtWidgets.QFrame()
            line.setFrameShape(QtWidgets.QFrame.HLine)
            line.setFrameShadow(QtWidgets.QFrame.Sunken)
            line.setLineWidth(1)
            return line

        self.surf_check = QtWidgets.QCheckBox("calculate surface based cavities", self)
        self.surf_check.setChecked(True)
        self.center_check = QtWidgets.QCheckBox("calculate center based cavities", self)
        self.gyration_tensor_check = QtWidgets.QCheckBox("calculate gyration tensor parameters", self)
        self.gyration_tensor_check.setToolTip("squared_gyration_radius, asphericity, acylindricity, anisotropy")
        self.overwrite_check = QtWidgets.QCheckBox("overwrite existing results", self)
        self.exporthdf5_check = QtWidgets.QCheckBox("export results as HDF5 files", self)
        self.exporttext_check = QtWidgets.QCheckBox("export results as text files", self)
        self.exportsingletext_check = QtWidgets.QCheckBox("export results as single text file", self)
        self.exportdir_radio_input = QtWidgets.QRadioButton("export to the directory of the input files", self)
        self.exportdir_radio_config = QtWidgets.QRadioButton("export to %s" % config.Path.result_dir, self)

        self.exportdir_radio_input.setChecked(True)

        self.exportdir_radio_group = QtWidgets.QButtonGroup(self)
        self.exportdir_radio_group.addButton(self.exportdir_radio_input)
        self.exportdir_radio_group.addButton(self.exportdir_radio_config)

        covalence_radii_by_element = self.__get_all_covalence_radii_by_element()
        self.radii_widget = RadiiWidget(covalence_radii_by_element, self.file_frame_dict, self)

        inner_layout.addLayout(res_hbox)
        inner_layout.addWidget(self.table_view)
        inner_layout.addWidget(self.surf_check)
        inner_layout.addWidget(self.center_check)
        inner_layout.addWidget(self.gyration_tensor_check)
        inner_layout.addWidget(create_separator())
        inner_layout.addWidget(self.overwrite_check)
        inner_layout.addWidget(self.exporthdf5_check)
        inner_layout.addWidget(self.exporttext_check)
        inner_layout.addWidget(self.exportsingletext_check)
        inner_layout.addWidget(create_separator())
        inner_layout.addWidget(self.exportdir_radio_input)
        inner_layout.addWidget(self.exportdir_radio_config)

        hbox.addLayout(inner_layout)
        hbox.addWidget(self.radii_widget)

        vbox.addLayout(hbox)
        vbox.addLayout(button_hbox)

        self.setLayout(vbox)

    def __get_all_covalence_radii_by_element(self):
        covalence_radii_by_element, elements_by_frame, element_combinations = self.__read_atom_info()
        return covalence_radii_by_element

    def __get_elements_by_frame(self):
        covalence_radii_by_element, elements_by_frame, element_combinations = self.__read_atom_info()
        return elements_by_frame

    def __get_element_combinations(self):
        covalence_radii_by_element, elements_by_frame, element_combinations = self.__read_atom_info()
        return element_combinations

    def __read_atom_info(self):
        method = self.__read_atom_info.__func__
        for attr in (
            "covalence_radii_by_element",
            "elements_by_frame",
            "element_combinations",
        ):
            if not hasattr(method, attr):
                setattr(method, attr, None)

        if (
            method.covalence_radii_by_element is None
            or method.elements_by_frame is None
            or method.element_combinations is None
        ):

            radii = {}
            elements_by_frame = {}
            element_combinations = set()
            for filepath, frames in self.file_frame_dict.items():
                elements_by_frame[filepath] = {}
                inputfile = file.File.open(filepath)
                if frames == (-1,):
                    frames = range(inputfile.info.num_frames)
                for frame in frames:
                    atoms = inputfile.getatoms(frame)
                    current_radii = atoms.covalence_radii_by_element
                    current_elements = atoms.elements
                    radii.update(current_radii)
                    elements_by_frame[filepath][frame] = current_elements
                    element_combinations.add(tuple(current_elements))
            method.covalence_radii_by_element = radii
            method.elements_by_frame = elements_by_frame
        return (
            method.covalence_radii_by_element,
            method.elements_by_frame,
            method.element_combinations,
        )

    def __update_cutoff_history(self):
        timestamp = datetime.datetime.now()
        user_cutoff_radii = self.radii_widget.cutoff_radii
        elements = self.__get_all_covalence_radii_by_element().keys()
        if not isinstance(user_cutoff_radii, collections.abc.Iterable):
            user_cutoff_radii = dict((elem, user_cutoff_radii) for elem in elements)
        elements_by_frame = self.__get_elements_by_frame()
        new_history = []
        for filepath, frames in self.file_frame_dict.items():
            filename = os.path.basename(filepath)
            file_elements = elements_by_frame[filepath]
            for frame in frames:
                elements = file_elements[frame]
                frame_cutoff_radii = dict((elem, user_cutoff_radii[elem]) for elem in elements)
                history_entry = HistoryEntry(filename, frame, timestamp, frame_cutoff_radii)
                new_history.append(history_entry)
        cutoff_history.extend(new_history)
        cutoff_history.save()

    def __update_cutoff_presets(self):
        save_preset_name = self.radii_widget.save_preset_name
        if save_preset_name is not None:
            user_cutoff_radii = self.radii_widget.cutoff_radii
            cutoff_presets.add(Preset(save_preset_name, user_cutoff_radii))
            cutoff_presets.save()

    def update_table(self):
        # get timestamps for selected frames for each file

        # surface based
        surface_ts = []
        center_ts = []
        for i, ts in enumerate(self.timestamps(center_based=False)):
            frames = (
                range(file.File.open(self.filenames[i]).info.num_frames)
                if self.file_frame_dict[self.filenames[i]][0] == -1
                else self.file_frame_dict[self.filenames[i]]
            )
            surface_ts.append([])
            for frame in frames:
                ts, temp = it.tee(ts)
                surface_ts[i].append(list(temp)[frame])

        # center based timestamps for the given frames
        center_ts = []
        for i, ts in enumerate(self.timestamps(center_based=True)):
            frames = (
                range(file.File.open(self.filenames[i]).info.num_frames)
                if self.file_frame_dict[self.filenames[i]][0] == -1
                else self.file_frame_dict[self.filenames[i]]
            )
            center_ts.append([])
            for frame in frames:
                ts, temp = it.tee(ts)
                center_ts[i].append(list(temp)[frame])

        # reduce to a single value per file
        surface_ts = ["X" if "X" in ts else ts[0] for ts in surface_ts]
        center_ts = ["X" if "X" in ts else ts[0] for ts in center_ts]
        basenames = [os.path.basename(path) for path in self.filenames]
        frames = [
            (
                str([frame + 1 for frame in self.file_frame_dict[f]])[1:-1]
                if not self.file_frame_dict[f][0] == -1
                else "all"
            )
            for f in self.filenames
        ]

        data_list = zip(basenames, surface_ts, center_ts, frames)

        # set table data
        header = ["dataset", "surface based", "center based", "frames"]
        table_model = TableModel(self, data_list, header)
        self.table_view.setModel(table_model)
        self.table_view.resizeColumnsToContents()

        # calculate table size to set its minimum size
        width = (self.table_view.model().columnCount(self.table_view) - 1) + self.table_view.verticalHeader().width()
        for i in range(self.table_view.model().columnCount(self.table_view)):
            width += self.table_view.columnWidth(i)
        self.table_view.setMinimumWidth(width)

        height = (self.table_view.model().rowCount(self.table_view) - 1) + self.table_view.horizontalHeader().height()
        for i in range(self.table_view.model().rowCount(self.table_view)):
            height += self.table_view.rowHeight(i)
        self.table_view.setMinimumHeight(height)

    def lineedit_return(self):
        try:
            self.resolution = int(self.lineedit.text())
            value = (self.resolution - self.RES_MIN) / self.RES_INTERVAL
            self.res_slider.setValue(value)
            self.update_table()
        except ValueError:
            pass

    def slider_changing(self, value):
        self.lineedit.setText(str(value * self.RES_INTERVAL + self.RES_MIN))

    def slider_released(self):
        self.resolution = self.res_slider.value() * self.RES_INTERVAL + self.RES_MIN
        self.update_table()

    def timestamps(self, center_based=False):
        return [
            self.control.calculation.calculatedframes(
                file.get_abspath(fn), self.resolution, not center_based, center_based
            ).prettystrings()
            for fn in self.filenames
        ]

    def ok(self):
        self.lineedit_return()
        self.__update_cutoff_history()
        self.__update_cutoff_presets()
        self.done(QtWidgets.QDialog.Accepted)

    def cancel(self):
        self.done(QtWidgets.QDialog.Rejected)

    def calculation_settings(self):
        calc_settings = None
        while True:
            ok = self.exec_()
            if ok != QtWidgets.QDialog.Accepted:
                break
            try:
                cutoff_radii = self.radii_widget.cutoff_radii
                surface_based = self.surf_check.isChecked()
                center_based = self.center_check.isChecked()
                gyration_tensor = self.gyration_tensor_check.isChecked()
                overwrite = self.overwrite_check.isChecked()
                exporthdf5 = self.exporthdf5_check.isChecked()
                exporttext = self.exporttext_check.isChecked()
                exportsingletext = self.exportsingletext_check.isChecked()
                if self.exportdir_radio_config.isChecked():
                    exportdir = os.path.expanduser(config.Path.result_dir)
                else:
                    exportdir = None
                calc_settings = calculation.CalculationSettings(
                    datasets=self.file_frame_dict,
                    resolution=self.resolution,
                    cutoff_radii=cutoff_radii,
                    domains=True,
                    surface_cavities=surface_based,
                    center_cavities=center_based,
                    gyration_tensor=gyration_tensor,
                    recalculate=overwrite,
                    exporthdf5=exporthdf5,
                    exporttext=exporttext,
                    exportsingletext=exportsingletext,
                    exportdir=exportdir,
                )
                break
            except ValueError:
                pass

        return (calc_settings, ok)


class RadiiWidget(QtWidgets.QWidget):
    class RadiiType:
        FIXED = 0
        CUSTOM = 1

    def __init__(self, radii, file_frame_dict, parent=None):
        super().__init__(parent)
        self._radii = radii
        self._file_frame_dict = file_frame_dict
        self._preferred_filenames_with_frames = dict(
            (os.path.basename(filepath), frames) for filepath, frames in self._file_frame_dict.items()
        )
        self._discard_preset_choice_on_next_rb_click = True  # Default value is True
        self._selected_radii_type = RadiiWidget.RadiiType.FIXED
        self._init_ui()
        self._init_cutoff_radii()

    def _init_ui(self):
        # Fixed Radio Button
        self.rb_fixed = QtWidgets.QRadioButton("Fixed Radius:")
        self.le_fixed = QtWidgets.QLineEdit()
        self.le_fixed.setMinimumWidth(150)
        self.le_fixed.setVisible(False)
        self.le_fixed.textEdited.connect(self.le_fixed_text_edited)
        self.rb_fixed.clicked.connect(self.rb_fixed_clicked)

        # QStackedWidget le_fixed
        self.tmp1 = QtWidgets.QWidget()
        self.sw_fixed = QtWidgets.QStackedWidget()
        self.sw_fixed.setSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Fixed)
        self.sw_fixed.addWidget(self.le_fixed)
        self.sw_fixed.addWidget(self.tmp1)
        self.sw_fixed.setCurrentIndex(1)

        # Custom Radio Button + Table
        self.rb_custom = QtWidgets.QRadioButton("Custom:")
        self.tw_cutoff = CutoffTableWidget(self._radii)
        self.tw_cutoff.text_edited.connect(self.tw_cutoff_text_edited)
        self.rb_custom.clicked.connect(self.rb_custom_clicked)

        # Preset Combo Box
        self.cb_preset = CutoffPresetComboBox()
        self.cb_preset.setSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Fixed)
        self.cb_preset.preset_selected.connect(self.cb_preset_selected)

        # History button
        self.pb_history = QtWidgets.QPushButton("History", self)
        self.pb_history.setMinimumWidth(0)
        self.pb_history.setSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Fixed)
        self.pb_history.setAutoDefault(False)
        self.pb_history.clicked.connect(self.pb_history_clicked)
        if (
            len(
                cutoff_history.filtered_history(
                    self._radii.keys(),
                    preferred_filenames_with_frames=self._preferred_filenames_with_frames,
                )
            )
            == 0
        ):
            self.pb_history.setVisible(False)

        # Preset save
        self.cb_preset_save = QtWidgets.QCheckBox("Save as Preset", self)
        self.le_preset_save = QtWidgets.QLineEdit()
        self.le_preset_save.setSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Fixed)
        self.le_preset_save.textChanged.connect(self.le_preset_save_changed)

        self.la_main = QtWidgets.QGridLayout()
        self.la_main.setContentsMargins(0, 0, 0, 0)
        self.la_fixed = QtWidgets.QHBoxLayout()
        self.la_fixed.setContentsMargins(0, 0, 0, 0)
        self.la_custom = QtWidgets.QHBoxLayout()
        self.la_custom.setContentsMargins(0, 0, 0, 0)
        self.la_preset_save = QtWidgets.QHBoxLayout()
        self.la_preset_save.setContentsMargins(0, 0, 0, 0)

        self.la_fixed.addWidget(self.rb_fixed)
        self.la_fixed.addWidget(self.sw_fixed)
        self.la_main.addLayout(self.la_fixed, 0, 0, 1, 2)
        self.la_custom.addWidget(self.rb_custom)
        self.la_main.addLayout(self.la_custom, 1, 0, 1, 2)
        self.la_main.addWidget(self.tw_cutoff, 2, 0, 1, 2)
        self.la_main.addWidget(self.cb_preset, 3, 0, 1, 1)
        self.la_main.addWidget(self.pb_history, 3, 1, 1, 1)
        self.la_preset_save.addWidget(self.cb_preset_save)
        self.la_preset_save.addWidget(self.le_preset_save)
        self.la_main.addLayout(self.la_preset_save, 4, 0, 1, 2)

        self.setLayout(self.la_main)

        self.rb_fixed.click()

    def _init_cutoff_radii(self, cutoff_radii=None):
        elements = self._radii.keys()
        if cutoff_radii is None:
            filtered_history = cutoff_history.filtered_history(
                elements,
                preferred_filenames_with_frames=self._preferred_filenames_with_frames,
            )
            if not filtered_history:
                cutoff_radii = config.Computation.std_cutoff_radius
            else:
                cutoff_radii = filtered_history[0].radii
        self.cutoff_radii = cutoff_radii

    def rb_fixed_clicked(self):
        self.sw_fixed.setCurrentIndex(0)
        if self._selected_radii_type != RadiiWidget.RadiiType.FIXED and self._discard_preset_choice_on_next_rb_click:
            self.cb_preset.discard_preset_choice()
            self._selected_radii_type = RadiiWidget.RadiiType.FIXED
        self._discard_preset_choice_on_next_rb_click = True

    def rb_custom_clicked(self):
        self.sw_fixed.setCurrentIndex(1)
        if self._selected_radii_type != RadiiWidget.RadiiType.CUSTOM and self._discard_preset_choice_on_next_rb_click:
            self.cb_preset.discard_preset_choice()
            self._selected_radii_type = RadiiWidget.RadiiType.CUSTOM
        self._discard_preset_choice_on_next_rb_click = True

    def le_fixed_text_edited(self):
        self.cb_preset.discard_preset_choice()

    def tw_cutoff_text_edited(self):
        self.cb_preset.discard_preset_choice()
        self.rb_custom.click()
        self.cb_preset.discard_preset_choice()

    def pb_history_clicked(self):
        cutoff_history_dialog = CutoffHistoryDialog(self, self._radii.keys(), self._preferred_filenames_with_frames)
        return_value = cutoff_history_dialog.exec_()
        if return_value == QtWidgets.QDialog.Rejected:
            return
        self._init_cutoff_radii(cutoff_history_dialog.selected_radii)

    def le_preset_save_changed(self, text):
        self.cb_preset_save.setChecked(text.strip() != "")

    def cb_preset_selected(self, selected_preset):
        self.cutoff_radii = selected_preset.radii

    @property
    def save_preset_name(self):
        if not self.cb_preset_save.isChecked():
            return None
        preset_name = self.le_preset_save.text().strip()
        if not re.match("[a-zA-Z0-9_]+", preset_name):
            message = QtWidgets.QMessageBox()
            message.setStandardButtons(QtWidgets.QMessageBox.Ok)
            message.setText(
                "{} is not valid preset name. Only A-Z, a-z, 0-9 and _ are valid characters".format(preset_name)
            )
            message.exec_()
            return None
        return preset_name

    @property
    def cutoff_radii(self):
        if self.rb_fixed.isChecked() and hasattr(self, "le_fixed"):
            try:
                radius_as_text = self.le_fixed.text()
                radius = float(radius_as_text)
                return radius
            except ValueError:
                message = QtWidgets.QMessageBox()
                message.setStandardButtons(QtWidgets.QMessageBox.Ok)
                message.setText("Incorrect input: {}".format(radius_as_text))
                message.exec_()
                raise

        if self.rb_custom.isChecked():
            elem_to_cutoff_radius = {}
            for i, elem in enumerate(self._radii):
                try:
                    current_radius_as_text = self.tw_cutoff.cellWidget(i, 1).text()
                    current_radius = float(current_radius_as_text)
                    elem_to_cutoff_radius[elem] = current_radius
                except ValueError:
                    message = QtWidgets.QMessageBox()
                    message.setStandardButtons(QtWidgets.QMessageBox.Ok)
                    message.setText("Incorrect input: {}".format(current_radius_as_text))
                    message.exec_()
                    raise
            return elem_to_cutoff_radius

    @cutoff_radii.setter
    def cutoff_radii(self, cutoff_radii):
        elements = self._radii.keys()
        if not isinstance(cutoff_radii, collections.abc.Iterable):
            cutoff_radii = dict((elem, float(cutoff_radii)) for elem in elements)
        # Check if all cutoff radii are equal
        if len(set(cutoff_radii.values())) == 1:
            self._discard_preset_choice_on_next_rb_click = False
            self.rb_fixed.click()
            self.le_fixed.setText(str(list(cutoff_radii.values())[0]))
        else:
            self._discard_preset_choice_on_next_rb_click = False
            self.rb_custom.click()
            for i, elem in enumerate(self._radii):
                self.tw_cutoff.cellWidget(i, 1).setText(str(cutoff_radii[elem]))



import os.path

from PySide6 import QtCore
from PySide6.QtWebEngineCore import QWebEnginePage  # QtWebKit is deprecated since Pyside2
from PySide6.QtWebEngineWidgets import QWebEngineView

_js_inject_css_template = """\
window.onload = function() {{
    (function(css) {{
        var node = document.createElement('style');
        node.innerHTML = css;
        document.body.appendChild(node);
    }})('{css}');
}};\
"""


class WebWidget(QWebEngineView):
    gui_link_clicked = QtCore.Signal(str)

    class WebPage(QWebEnginePage):
        gui_link_clicked = QtCore.Signal(str)

        def __init__(self, html, css=None, *args, **kwargs):
            super().__init__(*args, **kwargs)
            self._css = css
            self.setHtml(html)

        def setHtml(self, html):
            if html:
                css = " ".join(self._css.split("\n"))
                js = "<script>{code}</script>".format(code=_js_inject_css_template.format(css=css))
                html += js
            super(WebWidget.WebPage, self).setHtml(html)

        def acceptNavigationRequest(self, url, t, is_main_frame):
            if t == QWebEnginePage.NavigationTypeLinkClicked and url.scheme() == "gui":
                self.gui_link_clicked.emit(url.path())
                return False
            else:
                return True

    def __init__(self, css_filepath=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if css_filepath is None:
            self._css = None
        else:
            css_filepath = os.path.join(os.path.dirname(__file__), css_filepath)
            with open(css_filepath) as f:
                self._css = f.read()
        self._webpage = None

    def set_gui_html(self, html_string):
        if html_string is None:
            html_string = ""
        if self._webpage is None:
            self._webpage = WebWidget.WebPage(html_string, self._css, parent=self)
            self._webpage.gui_link_clicked.connect(self.gui_link_clicked.emit)  # Forward signal
        else:
            self._webpage.setHtml(html_string)
        self.setPage(self._webpage)



import csv
import sys
from math import floor

import gr
import numpy as np
from PySide6 import QtCore, QtWidgets
from qtgr import GRWidget

from ..util.logger import Logger

logger = Logger("gui.histogram_widget")
logger.setstream("default", sys.stdout, Logger.DEBUG)


class GrHistogramWidget(GRWidget):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.xvalues = None
        self.yvalues = None
        self.title = None
        self.datapoints = None

    def setdata(self, xvalues, yvalues, widths, title):
        self.xvalues = xvalues
        self.yvalues = yvalues
        self.widths = widths
        self.title = title

    def draw(self, wsviewport=None):
        if self.xvalues is not None and self.widths is not None:
            maxidx = np.argmax(self.xvalues)
            rangex = (self.xvalues.min(), self.xvalues[maxidx] + self.widths[maxidx])
        else:
            rangex = (0.0, 100.0)
        if self.yvalues is not None:
            rangey = gr.adjustrange(0.0, self.yvalues.max())
        else:
            rangey = (0.0, 8.0)

        if wsviewport is None:
            gr.setwsviewport(0, self.mwidth, 0, self.mheight)
        else:
            gr.setwsviewport(*wsviewport)
        gr.setwswindow(0, self.sizex, 0, self.sizey)
        gr.setviewport(0.075 * self.sizex, 0.95 * self.sizex, 0.075 * self.sizey, 0.95 * self.sizey)
        gr.setwindow(rangex[0], rangex[1], rangey[0], rangey[1])
        gr.setcharheight(0.012)

        gr.setfillintstyle(1)
        gr.setfillcolorind(0)
        gr.fillrect(rangex[0], rangex[1], rangey[0], rangey[1])

        if self.xvalues is not None and self.yvalues is not None and self.widths is not None:
            gr.setfillintstyle(1)
            gr.setfillcolorind(2)
            for i in range(self.xvalues.size):
                gr.fillrect(
                    self.xvalues[i],
                    self.xvalues[i] + self.widths[i] * 0.8,
                    0.0,
                    self.yvalues[i],
                )
        else:
            gr.text(0.45 * self.sizex, 0.5 * self.sizey, "no data")

        gr.setlinecolorind(1)
        xtick = floor(0.02 * (rangex[1] - rangey[0]) * 100.0) / 100.0
        ytick = floor(0.04 * (rangey[1] - rangey[0]) * 50.0) / 50.0
        gr.axes(xtick, ytick, rangex[0], rangey[0], 10, 5, 0.0075)
        gr.axes(xtick, ytick, rangex[1], rangey[1], -10, -5, -0.0075)

        if self.title is not None:
            gr.text(0.8 * self.sizex, 0.9 * self.sizey, self.title)


class HistogramWidget(QtWidgets.QWidget):

    def __init__(self, parent):
        super().__init__(parent)

        self.control = parent.control
        self.results = None

        self.init_gui()

    def init_gui(self):
        vbox = QtWidgets.QVBoxLayout()
        grid = QtWidgets.QGridLayout()
        self.gr_widget = GrHistogramWidget()

        self.datasetlabel = QtWidgets.QLabel("No data loaded.", self)
        self.datasetlabel.setAlignment(QtCore.Qt.AlignHCenter)

        selectbox = QtWidgets.QHBoxLayout()
        self.cavity_type_box = QtWidgets.QComboBox(self)
        self.cavity_type_box.setMinimumWidth(180)
        selectbox.addWidget(self.cavity_type_box)
        selectbuttongroup = QtWidgets.QButtonGroup(self)
        self.volumebutton = QtWidgets.QRadioButton("Cavity Volume", self)
        selectbox.addWidget(self.volumebutton)
        selectbuttongroup.addButton(self.volumebutton)
        self.areabutton = QtWidgets.QRadioButton("Surface Area", self)
        selectbox.addWidget(self.areabutton)
        selectbuttongroup.addButton(self.areabutton)
        self.volumebutton.setChecked(True)
        grid.addLayout(selectbox, 0, 0)

        self.weightbutton = QtWidgets.QCheckBox("Weighted Histogram", self)
        self.weightbutton.setChecked(True)
        grid.addWidget(self.weightbutton, 0, 1)

        binbox = QtWidgets.QHBoxLayout()
        binbox.addWidget(QtWidgets.QLabel("Number of Bins:", self), 0)
        self.nbins = QtWidgets.QLineEdit(self)
        self.nbins.setMinimumWidth(50)
        binbox.addWidget(self.nbins, 0, QtCore.Qt.AlignLeft)
        grid.addLayout(binbox, 0, 2)

        buttonbox = QtWidgets.QHBoxLayout()

        self.plotbutton = QtWidgets.QPushButton("Plot", self)
        buttonbox.addWidget(self.plotbutton)
        self.plotbutton.clicked.connect(self.draw)

        self.export_image_button = QtWidgets.QPushButton("Save Image", self)
        buttonbox.addWidget(self.export_image_button)
        self.export_image_button.clicked.connect(self.export_image)

        self.export_data_button = QtWidgets.QPushButton("Export Data", self)
        buttonbox.addWidget(self.export_data_button)
        self.export_data_button.clicked.connect(self.export_data)
        grid.addLayout(buttonbox, 1, 0, 1, 3)

        vbox.addWidget(self.gr_widget, stretch=1)
        vbox.addWidget(self.datasetlabel, stretch=0)
        vbox.addLayout(grid)
        self.setLayout(vbox)
        self.show()

    def keyPressEvent(self, event):
        if event.key() == QtCore.Qt.Key_Return or event.key() == QtCore.Qt.Key_Enter:
            self.update()

    def draw(self, now=False, wsviewport=None):
        xvalues = None
        yvalues = None
        widths = None
        title = None
        if self.results is None:
            self.refresh()
        if self.cavity_type_box.currentText():
            if self.cavity_type_box.currentText() == "Surface-based Cavities":
                cavities = self.results.surface_cavities
            elif self.cavity_type_box.currentText() == "Center-based Cavities":
                cavities = self.results.center_cavities
            elif self.cavity_type_box.currentText() == "Cavity Domains":
                cavities = self.results.domains
            nbins = str(self.nbins.text())
            if len(nbins) > 0 and int(nbins) > 0:
                nbins = int(nbins)
            else:
                nbins = 50
            if self.volumebutton.isChecked():
                data = cavities.volumes
            else:
                data = cavities.surface_areas
            if len(data) > 0:
                if self.weightbutton.isChecked():
                    hist, bin_edges = np.histogram(data, bins=nbins, weights=data)
                else:
                    hist, bin_edges = np.histogram(data, bins=nbins)
                widths = bin_edges[1:] - bin_edges[:-1]
                xvalues = bin_edges[:-1]
                yvalues = hist

        self.gr_widget.setdata(xvalues, yvalues, widths, title)
        self.gr_widget.update()
        if now:
            self.gr_widget.draw(wsviewport=wsviewport)

    def export_image(self):
        extensions = (
            ".pdf",
            ".png",
            ".bmp",
            ".jpg",
            ".jpeg",
            ".png",
            ".tiff",
            ".fig",
            ".svg",
            ".wmf",
            ".eps",
            ".ps",
        )
        qtext = "*" + " *".join(extensions)
        filepath = QtWidgets.QFileDialog.getSaveFileName(self, "Save Image", ".", "Image Files ({})".format(qtext))[0]
        if not filepath:
            return

        if filepath.endswith(".eps") or filepath.endswith(".ps"):
            gr.beginprintext(filepath, "Color", "A4", "Landscape")
            self.draw(now=True, wsviewport=(0, 0.297 * 0.9, 0, 0.21 * 0.95))
        else:
            gr.beginprint(filepath)
            self.draw(now=True)
        gr.endprint()

    def export_data(self):
        qtext = " *.csv"
        filepath = QtWidgets.QFileDialog.getSaveFileName(self, "Save Data", ".", "CSV Files ({})".format(qtext))[0]
        if len(filepath) == 0:
            return
        self.update()
        xvalues = self.gr_widget.xvalues
        yvalues = self.gr_widget.yvalues
        if xvalues is None or yvalues is None:
            return
        with open(filepath, "w") as csvfile:
            csvwriter = csv.writer(csvfile)
            for x, y in zip(xvalues, yvalues):
                csvwriter.writerow([x, y])

    def refresh(self):
        self.cavity_type_box.clear()
        results = self.control.results
        if results is not None:
            results = results[-1][-1]
            if results.surface_cavities is not None:
                self.cavity_type_box.addItem("Surface-based Cavities")
            if results.center_cavities is not None:
                self.cavity_type_box.addItem("Center-based Cavities")
            if results.domains is not None:
                self.cavity_type_box.addItem("Cavity Domains")
            if self.results != results:
                self.results = results
                self.gr_widget.setdata(None, None, None, None)
                self.gr_widget.update()
            self.datasetlabel.setText(str(results))
        else:
            self.datasetlabel.setText("")

    def activate(self):
        self.refresh()

    def updatestatus(self):
        self.refresh()



"""
Visualize Atoms and Cavities with GR3
"""

from ctypes import c_int

import gr3
import numpy as np
import numpy.linalg as la

from ..config.configuration import config
from ..util.gl_util import (
    create_look_at_matrix,
    create_perspective_projection_matrix,
    create_rotation_matrix_homogenous,
    create_translation_matrix_homogenous,
)


class Visualization(object):
    """
    Visualize Atoms and Cavities with GR3
    """

    def __init__(self):
        self.max_side_lengths = 1.0
        self.mat = np.eye(4)
        self.d = self.max_side_lengths * 2
        self.pos = np.array((0, 0, self.d))
        self.up = np.array((0, 1, 0))
        self.right = np.array((1, 0, 0))

        self.near = 0.1
        self.far = 3 * self.d
        self.fov = 45.0

        self.width = 0
        self.height = 0
        self.usecurrentframebuffer = False
        #  self.assigned_opengl_context = None

        self.results = None
        self.settings = VisualizationSettings()
        self.objectids = [None]

    def setresults(self, results):
        """
        Set the results that will be displayed

        **Parameters:**
            `results` :
                :class:`core.data.Results` object to visualize
        """
        self.results = results
        max_side_lengths = max(results.atoms.volume.side_lengths)
        self.d = self.d / self.max_side_lengths * max_side_lengths
        self.max_side_lengths = max_side_lengths
        self.far = 6 * max_side_lengths
        self.create_scene()
        if self.width != 0 and self.height != 0:
            self.paint(self.width, self.height)
        self.settings.visible_domain_indices = None
        self.settings.visible_surface_cavity_indices = None
        self.settings.visible_center_cavity_indices = None

    def create_scene(self):
        """
        Create GR3 meshes. ``self.results`` contains the mesh data
        and in ``self.settings`` is specified which meshes shound be rendered.
        """

        c = config.Colors.background
        gr3.setbackgroundcolor(c[0], c[1], c[2], 1.0)
        gr3.clear()

        if self.results is None:
            return

        show_domains = self.settings.show_domains
        show_surface_cavities = self.settings.show_surface_cavities
        show_center_cavities = self.settings.show_center_cavities
        if show_center_cavities and self.results.center_cavities is not None:
            show_surface_cavities = False
        elif show_surface_cavities and self.results.surface_cavities is not None:
            show_domains = False

        self.objectids = [None]
        edges = self.results.atoms.volume.edges
        num_edges = len(edges)
        edge_positions = [edge[0] for edge in edges]
        edge_directions = [[edge[1][i] - edge[0][i] for i in range(3)] for edge in edges]
        edge_lengths = [sum([c * c for c in edge]) ** 0.5 for edge in edge_directions]
        edge_radius = min(edge_lengths) / 200
        if self.settings.show_bounding_box:
            gr3.drawcylindermesh(
                num_edges,
                edge_positions,
                edge_directions,
                [config.Colors.bounding_box] * num_edges,
                [edge_radius] * num_edges,
                edge_lengths,
            )
            corners = list(set([tuple(edge[0]) for edge in edges] + [tuple(edge[1]) for edge in edges]))
            num_corners = len(corners)
            gr3.drawspheremesh(
                num_corners,
                corners,
                [config.Colors.bounding_box] * num_corners,
                [edge_radius] * num_corners,
            )

        if self.settings.show_atoms and self.results.atoms is not None:
            visible_atom_indices = self.settings.visible_atom_indices
            if visible_atom_indices is not None:
                visible_atom_indices = [comp for comp in visible_atom_indices if 0 <= comp < self.results.atoms.number]
            else:
                visible_atom_indices = range(self.results.atoms.number)
            if len(visible_atom_indices) == 0:
                visible_atom_indices = None
            if visible_atom_indices is not None:
                visible_atom_indices = np.array(visible_atom_indices)
                gr3.drawspheremesh(
                    len(visible_atom_indices),
                    self.results.atoms.positions[visible_atom_indices],
                    self.results.atoms.colors[visible_atom_indices],
                    np.ones(len(visible_atom_indices)) * config.OpenGL.atom_radius,
                )
                if self.settings.show_bonds:
                    bonds = self.results.atoms.bonds
                    for start_index, target_indices in enumerate(bonds):
                        if start_index not in visible_atom_indices:
                            continue
                        target_indices = np.array([i for i in target_indices if i in visible_atom_indices])
                        if len(target_indices) == 0:
                            continue
                        start_position = self.results.atoms.positions[start_index]
                        target_positions = self.results.atoms.positions[target_indices]
                        directions = target_positions - start_position
                        bond_lengths = la.norm(directions, axis=1)
                        directions /= bond_lengths.reshape(len(directions), 1)
                        gr3.drawcylindermesh(
                            len(target_indices),
                            target_positions,
                            -directions,
                            [config.Colors.bonds] * self.results.atoms.number,
                            np.ones(bond_lengths.shape) * config.OpenGL.bond_radius,
                            bond_lengths,
                        )

        if self.results is None:
            return
        if show_domains and self.results.domains is not None:
            self.draw_cavities(
                self.results.domains,
                config.Colors.domain,
                "domain",
                self.settings.visible_domain_indices,
            )
        if show_surface_cavities and self.results.surface_cavities is not None:
            self.draw_cavities(
                self.results.surface_cavities,
                config.Colors.surface_cavity,
                "surface cavity",
                self.settings.visible_surface_cavity_indices,
            )
        if show_center_cavities and self.results.center_cavities is not None:
            self.draw_cavities(
                self.results.center_cavities,
                config.Colors.center_cavity,
                "center cavity",
                self.settings.visible_center_cavity_indices,
            )

    def draw_cavities(self, cavities, color, cavity_type, indices=None):
        if indices is None:
            indices = range(self.results.domains.number)
        for index in set(indices):
            if 0 <= index < len(cavities.triangles):
                triangles = cavities.triangles[index]
                gr3._gr3.gr3_setobjectid(gr3.c_int(len(self.objectids)))
                self.objectids.append((cavity_type, index))
                mesh = gr3.createmesh(
                    triangles.shape[1] * 3,
                    triangles[0, :, :, :],
                    triangles[1, :, :, :],
                    [color] * (triangles.shape[1] * 3),
                )
                gr3.drawmesh(mesh, 1, (0, 0, 0), (0, 0, 1), (0, 1, 0), (1, 1, 1), (1, 1, 1))
                gr3.deletemesh(c_int(mesh.value))
        gr3._gr3.gr3_setobjectid(gr3.c_int(0))

    def zoom(self, delta):
        """
        Adjust the zoom level.

        **Parameters:**
            `delta` :
                the increment of the distance from the camera to the center
        """
        zoom_v = 1.0 / 20
        zoom_cap = self.d + zoom_v * delta < (self.max_side_lengths) * 4
        if self.d + zoom_v * delta > 0 and zoom_cap:
            self.d += zoom_v * delta

    def translate_mouse(self, dx, dy):
        """
        Translate the model.
        """
        trans_v = 1.0 / 1000
        trans_factor = trans_v * max(self.d, 20)
        m = create_translation_matrix_homogenous(-dx * trans_factor, dy * trans_factor, 0)
        self.mat = self.mat.dot(m)

    def set_focus_on(self, x, y, z):
        self.mat[:3, 3] = (x, y, z)

    def rotate_mouse(self, dx, dy):
        """
        Rotate the model according to a mouse movement (dx, dy) on the screen.
        """
        rot_v = 1.0 / 13000
        if dx == 0 and dy == 0:
            return
        rot_axis = np.array((-dy, -dx, 0.0))
        rot_factor = max(self.d, 20) * rot_v * la.norm(rot_axis)
        # rotation matrix with min rotation angle
        m = create_rotation_matrix_homogenous(rot_factor, rot_axis[0], rot_axis[1], rot_axis[2])
        self.mat = self.mat.dot(m)

    def reset_view(self):
        self.mat = np.eye(4)

    def set_camera(self, width, height):
        """
        Update the shown scene after the perspective has changed.
        """
        # rightt = self.mat[:3, 0]
        upt = self.mat[:3, 1]
        pt = self.mat[:3, 2] * self.d
        t = self.mat[:3, 3]

        self.proj_mat = create_perspective_projection_matrix(
            np.radians(self.fov), 1.0 * width / height, self.near, self.far
        )
        gr3.setcameraprojectionparameters(self.fov, self.near, self.far)
        self.lookat_mat = create_look_at_matrix(pt + t, t, upt)
        gr3.cameralookat(
            pt[0] + t[0],
            pt[1] + t[1],
            pt[2] + t[2],
            t[0],
            t[1],
            t[2],
            upt[0],
            upt[1],
            upt[2],
        )

    def assign_opengl_context(self, context):
        self.assigned_opengl_context = context

    def paint(self, width, height, usecurrentframebuffer=None, device_pixel_ratio=1):
        """
        Refresh the OpenGL scene.
        """
        self.width = width
        self.height = height
        self.set_camera(width, height)
        if usecurrentframebuffer is not None:
            self.usecurrentframebuffer = usecurrentframebuffer
        if self.usecurrentframebuffer:
            gr3.usecurrentframebuffer()
        gr3.drawimage(
            0,
            width * device_pixel_ratio,
            0,
            height * device_pixel_ratio,
            width,
            height,
            gr3.GR3_Drawable.GR3_DRAWABLE_OPENGL,
        )

    def save_screenshot(self, file_name, width=3840, height=2160, first=True, last=True):
        """
        Save a screenshot in the given resolution.
        ``first`` and ``last`` can be used to indicate if the first or last screenshot is taken when
        multiple images are saved in a loop for example.
        """
        gr3.export(file_name, width, height)

    def get_object_at_2dposition(self, x, y):
        oid = gr3.c_int(0)
        gr3._gr3.gr3_selectid(
            gr3.c_int(x),
            gr3.c_int(y),
            gr3.c_int(self.width),
            gr3.c_int(self.height),
            gr3.byref(oid),
        )
        return self.objectids[oid.value]

    def get_object_at_3dposition(self, x, y, z):
        if self.results is None:
            return None
        if self.results.atoms is not None:
            positions = self.results.atoms.positions
            distances = la.norm(positions - (x, y, z), axis=1)
            nearest_atom_index = np.argmin(distances)

            # calculate atom radius
            edges = self.results.atoms.volume.edges
            edge_directions = [[edge[1][i] - edge[0][i] for i in range(3)] for edge in edges]
            edge_lengths = [sum([c * c for c in edge]) ** 0.5 for edge in edge_directions]
            edge_radius = min(edge_lengths) / 200
            atom_radius = 4 * edge_radius

            if 0.95 < distances[nearest_atom_index] / atom_radius < 1.05:
                return ("atom", nearest_atom_index)


class VisualizationSettings(object):
    """
    Container for settings that control which part of the data is visualized.
    Attributes:

        `show_domains`

        `show_surface_cavities`

        `show_center_cavities`

        `show_atoms`

        `show_bounding_box`

    """

    def __init__(
        self,
        domains=False,
        show_surface_cavities=True,
        show_center_cavities=False,
        atoms=True,
        bonds=True,
        bounding_box=True,
        visible_atom_indices=None,
        visible_domain_indices=None,
        visible_surface_cavity_indices=None,
        visible_center_cavity_indices=None,
    ):
        self.show_domains = domains
        self.visible_domain_indices = visible_domain_indices
        self.show_surface_cavities = show_surface_cavities
        self.visible_surface_cavity_indices = visible_surface_cavity_indices
        self.show_center_cavities = show_center_cavities
        self.visible_center_cavity_indices = visible_center_cavity_indices
        self.show_atoms = atoms
        self.visible_atom_indices = visible_atom_indices
        self.show_bonds = bonds
        self.show_bounding_box = bounding_box



import csv
import sys

import gr
import numpy as np
from PySide6 import QtCore, QtWidgets
from qtgr import GRWidget

from ..statistics.pdf import PDF, Kernels
from ..util.logger import Logger

logger = Logger("gui.pdf_widget")
logger.setstream("default", sys.stdout, Logger.DEBUG)


class GrPlotWidget(GRWidget):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.xvalues = None
        self.yvalues = None
        self.title = None
        self.datapoints = None

    def quit(self):
        gr.emergencyclosegks()
        self.close()

    def setdata(self, xvalues, yvalues, title, datapoints=None):
        self.xvalues = xvalues
        self.yvalues = yvalues
        self.title = title
        self.datapoints = datapoints

    def draw(self, wsviewport=None):
        if self.xvalues is not None:
            rangex = (self.xvalues.min(), self.xvalues.max())
        else:
            rangex = (0, 10)
        if self.yvalues is not None:
            rangey = gr.adjustrange(self.yvalues.min(), self.yvalues.max())
        else:
            rangey = (0, 4)

        if wsviewport is None:
            gr.setwsviewport(0, self.mwidth, 0, self.mheight)
        else:
            gr.setwsviewport(*wsviewport)
        gr.setwswindow(0, self.sizex, 0, self.sizey)
        gr.setviewport(0.075 * self.sizex, 0.95 * self.sizex, 0.075 * self.sizey, 0.95 * self.sizey)
        gr.setwindow(rangex[0], rangex[1], rangey[0], rangey[1])
        gr.setcharheight(0.012)

        gr.setfillintstyle(1)
        gr.setfillcolorind(0)
        gr.fillrect(rangex[0], rangex[1], rangey[0], rangey[1])

        if self.xvalues is not None and self.yvalues is not None:
            gr.setlinecolorind(2)
            gr.polyline(self.xvalues, self.yvalues)
        else:
            gr.text(0.4 * self.sizex, 0.5 * self.sizey, "no elements selected")

        gr.setlinecolorind(1)
        gr.axes(0.2, 0.2, rangex[0], rangey[0], 5, 5, 0.0075)
        gr.axes(0.2, 0.2, rangex[1], rangey[1], -5, -5, -0.0075)

        if self.title is not None:
            gr.text(0.8 * self.sizex, 0.9 * self.sizey, self.title)


class PDFWidget(QtWidgets.QWidget):

    def __init__(self, parent):
        super().__init__(parent)

        self.control = parent.control
        self.results = None
        self.pdf = None

        self.init_gui()

    def init_gui(self):
        vbox = QtWidgets.QVBoxLayout()
        grid = QtWidgets.QGridLayout()
        self.gr_widget = GrPlotWidget()

        self.datasetlabel = QtWidgets.QLabel("No data loaded.", self)
        self.datasetlabel.setAlignment(QtCore.Qt.AlignHCenter)

        elembox = QtWidgets.QHBoxLayout()
        elembox.addWidget(QtWidgets.QLabel("Elements:", self), 0)
        self.elem1 = QtWidgets.QComboBox(self)
        self.elem1.setMinimumWidth(170)
        elembox.addWidget(self.elem1, 1)
        self.elem2 = QtWidgets.QComboBox(self)
        self.elem2.setMinimumWidth(170)
        elembox.addWidget(self.elem2, 1)
        grid.addLayout(elembox, 0, 0)

        rangebox = QtWidgets.QHBoxLayout()
        rangebox.addWidget(QtWidgets.QLabel("Plot range:", self))
        self.range1 = QtWidgets.QLineEdit("0", self)
        self.range1.setMinimumWidth(30)
        rangebox.addWidget(self.range1)
        rangebox.addWidget(QtWidgets.QLabel("-", self))
        self.range2 = QtWidgets.QLineEdit("8", self)
        self.range2.setMinimumWidth(30)
        rangebox.addWidget(self.range2)
        grid.addLayout(rangebox, 1, 0)

        cutoffbox = QtWidgets.QHBoxLayout()
        cutoffbox.addWidget(QtWidgets.QLabel("Kernel:", self))
        self.kernels = {
            "Gaussian": Kernels.gauss,
            "Epanechnikov": Kernels.epanechnikov,
            "Compact": Kernels.compact,
            "Triangular": Kernels.triang,
            "Box": Kernels.quad,
            "Right Box": Kernels.posquad,
            "Left Box": Kernels.negquad,
        }
        self.kernel = QtWidgets.QComboBox(self)
        self.kernel.setMinimumWidth(130)
        self.kernel.addItems(
            [
                "Gaussian",
                "Epanechnikov",
                "Compact",
                "Triangular",
                "Box",
                "Right Box",
                "Left Box",
            ]
        )
        cutoffbox.addWidget(self.kernel)
        cutoffbox.addWidget(QtWidgets.QLabel("Cutoff:", self))
        self.cutoff = QtWidgets.QLineEdit("12", self)
        self.cutoff.setMinimumWidth(30)
        cutoffbox.addWidget(self.cutoff)
        cutoffbox.addWidget(QtWidgets.QLabel("Bandwidth:", self))
        self.bandwidth = QtWidgets.QLineEdit("", self)
        self.bandwidth.setMinimumWidth(30)
        cutoffbox.addWidget(self.bandwidth)
        grid.addLayout(cutoffbox, 2, 0)

        buttonbox = QtWidgets.QHBoxLayout()

        self.plotbutton = QtWidgets.QPushButton("Plot", self)
        buttonbox.addWidget(self.plotbutton)
        self.plotbutton.clicked.connect(self.draw)

        self.export_image_button = QtWidgets.QPushButton("Save Image", self)
        buttonbox.addWidget(self.export_image_button)
        self.export_image_button.clicked.connect(self.export_image)

        self.export_data_button = QtWidgets.QPushButton("Export Data", self)
        buttonbox.addWidget(self.export_data_button)
        self.export_data_button.clicked.connect(self.export_data)
        grid.addLayout(buttonbox, 3, 0)

        vbox.addWidget(self.gr_widget, stretch=1)
        vbox.addWidget(self.datasetlabel, stretch=0)
        vbox.addLayout(grid)
        self.setLayout(vbox)
        self.show()

    def keyPressEvent(self, event):
        if event.key() == QtCore.Qt.Key_Return or event.key() == QtCore.Qt.Key_Enter:
            self.update()

    def draw(self, now=False, wsviewport=None):
        xvalues = None
        yvalues = None
        title = None
        datapoints = None
        if self.pdf is None:
            self.refresh()
        if self.pdf is not None:
            elem1 = str(self.elem1.currentText())
            if elem1 == "cavity domain centers":
                elem1 = "cav"
            elem2 = str(self.elem2.currentText())
            if elem2 == "cavity domain centers":
                elem2 = "cav"
            range1 = float(str(self.range1.text()))
            range2 = float(str(self.range2.text()))
            cutoff = str(self.cutoff.text())
            if len(cutoff) > 0 and float(cutoff) > 0:
                cutoff = float(cutoff)
            else:
                cutoff = None
            try:
                bandwidth = float(str(self.bandwidth.text()))
                if bandwidth < 0:
                    bandwidth = 0
                    self.bandwidth.setText("0")
            except ValueError:
                bandwidth = None
                self.bandwidth.setText("")
            kernel = self.kernels.get(self.kernel.currentText(), None)
            f = self.pdf.pdf(elem1, elem2, cutoff=cutoff, h=bandwidth, kernel=kernel)
            if f is not None:
                if callable(f):
                    xvalues = np.linspace(range1, range2, 400)
                    yvalues = f(xvalues)
                    title = "{} - {}".format(elem1, elem2)
                    datapoints = f.f.x
                else:
                    peaks = f[np.logical_and(range1 < f, f < range2)]
                    if len(peaks) > 2:
                        xvalues = np.zeros(len(peaks) * 3)
                        xvalues[0::3] = peaks
                        xvalues[1::3] = peaks
                        xvalues[2::3] = peaks
                        yvalues = np.zeros(len(peaks) * 3)
                        yvalues[1::3] = 1
                        datapoints = peaks

        self.gr_widget.setdata(xvalues, yvalues, title, datapoints)
        self.gr_widget.update()
        if now:
            self.gr_widget.draw(wsviewport=wsviewport)

    def export_image(self):
        extensions = (
            ".pdf",
            ".png",
            ".bmp",
            ".jpg",
            ".jpeg",
            ".png",
            ".tiff",
            ".fig",
            ".svg",
            ".wmf",
            ".eps",
            ".ps",
        )
        qtext = "*" + " *".join(extensions)
        filepath = QtWidgets.QFileDialog.getSaveFileName(self, "Save Image", ".", "Image Files ({})".format(qtext))[0]
        if len(filepath) == 0:
            return

        if filepath.endswith(".eps") or filepath.endswith(".ps"):
            gr.beginprintext(filepath, "Color", "A4", "Landscape")
            self.draw(now=True, wsviewport=(0, 0.297 * 0.9, 0, 0.21 * 0.95))
        else:
            gr.beginprint(filepath)
            self.draw(now=True)
        gr.endprint()

    def export_data(self):
        qtext = " *.csv"
        filepath = QtWidgets.QFileDialog.getSaveFileName(self, "Save Data", ".", "CSV Files ({})".format(qtext))[0]
        if len(filepath) == 0:
            return
        self.update()
        xvalues = self.gr_widget.xvalues
        yvalues = self.gr_widget.yvalues
        if xvalues is None or yvalues is None:
            return
        with open(filepath, "w") as csvfile:
            csvwriter = csv.writer(csvfile)
            for x, y in zip(xvalues, yvalues):
                csvwriter.writerow([x, y])

    def refresh(self):
        results = self.control.results
        if results is not None:
            results = results[-1][-1]
            if self.results != results or self.pdf is None:
                self.results = results
                self.pdf = PDF(results)
                e = np.unique(results.atoms.elements).tolist()
                for i in range(len(e)):
                    e[i] = e[i].decode("utf-8")
                if results.domains is not None and len(results.domains.centers) > 0 and "cav" not in e:
                    e.append("cavity domain centers")
                self.elem1.clear()
                self.elem1.addItems(e)
                self.elem2.clear()
                self.elem2.addItems(e)
                self.gr_widget.setdata(None, None, None)
                self.gr_widget.update()
            self.datasetlabel.setText(str(results))
        else:
            self.datasetlabel.setText("")
            self.elem1.clear()
            self.elem2.clear()

    def activate(self):
        self.refresh()

    def updatestatus(self):
        self.refresh()



"""
This module provides classes to handle pyMolDyn related files.
"""

import os
import os.path
import sys
from itertools import repeat

import h5py

from .. import core
from ..util.logger import Logger
from . import data

try:
    import pybel

    raise ImportError
except ImportError:

    class pybel:
        """
        Dummy class representing a missing pybel module.
        """

        informats = {}


__all__ = ["File", "InputFile", "ResultFile", "XYZFile", "HDF5File"]

logger = Logger("core.file")
logger.setstream("default", sys.stdout, Logger.WARNING)

SEARCH_PATH = None


def get_abspath(path):
    """
    Return an absolute path for an input file. The difference to abspath from
    os.path is that this function uses SEARCH_PATH instead of the current
    working directory
    """
    global SEARCH_PATH
    if not os.path.isabs(path) and SEARCH_PATH is not None:
        # relative paths use the SEARCH_PATH instead of the current working
        # directory, as that is modified at startup. The SEARCH_PATH is set
        # to the initial cwd in startBatch and startGUI.
        path = os.path.join(SEARCH_PATH, path)
    return os.path.abspath(path)


class FileError(Exception):
    """
    Exception which will be raised by the subclasses of :class:`InputFile`
    if errors occur when handling the data inside the file.
    """

    def __init__(self, message, cause=None):
        super().__init__(message, cause)
        self.message = message
        self.cause = cause

    def __str__(self):
        if self.cause is None:
            f = "{}: {}"
        else:
            f = "{}: {} (caused by: {})"
        return f.format(self.__class__.__name__, self.message, str(self.cause))


class InputFile(object):
    """
    Abstract access to a file that contains atom data for one or more frames.
    Subclasses need to implement the ``readinfo()`` and ``readatoms(frame)`` methods.
    """

    def __init__(self, path):
        """
        **Parameters:**
            `path` :
                absolute path to the file
        """
        self.path = get_abspath(path)
        self._info = data.FileInfo()
        self.inforead = False

    @property
    def info(self):
        """
        :class:`core.data.FileInfo` object that contains metadata
        """
        if not self.inforead:
            try:
                self.readinfo()
            except IOError as e:
                logger.info(str(e))
                return None
            if self._info.volume is None:
                self._info.volume_guessed = True
                minx, maxx = float("inf"), float("-inf")
                miny, maxy = float("inf"), float("-inf")
                minz, maxz = float("inf"), float("-inf")
                for frame in range(self._info.num_frames):
                    atoms = self.getatoms(frame)
                    minx = min(minx, atoms.positions[:, 0].min())
                    maxx = max(maxx, atoms.positions[:, 0].max())
                    miny = min(miny, atoms.positions[:, 1].min())
                    maxy = max(maxy, atoms.positions[:, 1].max())
                    minz = min(minz, atoms.positions[:, 2].min())
                    maxz = max(maxz, atoms.positions[:, 2].max())
                self._info.volumestr = "ORT %f %f %f" % (
                    maxx - minx,
                    maxy - miny,
                    maxz - minz,
                )
        return self._info

    def getatoms(self, frame):
        """
        Read atom data for a specified frame.

        **Parameters:**
            `frame` :
                the frame number

        **Returns:**
            an :class:`core.data.Atoms` object

        **Raises:**
            - :class:`IndexError`: if the frame is not in the file
            - :class:`FileError`: if there are problems with the data in the file
            - :class:`IOError`: if the file cannot be read
        """
        return self.readatoms(frame)

    def readinfo(self):
        raise NotImplementedError

    def readatoms(self, frame):
        raise NotImplementedError


class XYZFile(InputFile):
    """
    Implementation on :class:`InputFile` for Open Babel 'xyz' files.
    """

    def __init__(self, path):
        super().__init__(path)
        f = open(self.path, "r")
        f.close()

    def readinfo(self):
        try:
            self._info.num_frames = 0
            with open(self.path.encode("utf-8"), "r") as f:
                try:
                    while True:
                        num_atoms = f.readline()
                        if not num_atoms:
                            break
                        num_atoms = int(num_atoms)
                        volume_info = f.readline()
                        for i in range(num_atoms):
                            f.readline()
                        self._info.num_frames += 1
                        if self._info.num_frames == 1:
                            self._info.volumestr = volume_info
                except StopIteration:
                    pass
            self.inforead = True
        except IOError:
            raise
        except Exception as e:
            raise FileError("Cannot read file info.", e)

    def readatoms(self, frame):
        try:
            if self.info.num_frames <= frame:
                raise IndexError("Frame {} not found".format(frame))
            with open(self.path.encode("utf-8"), "r") as f:
                try:
                    # Skip the first frames
                    for i in range(frame):
                        num_atoms = f.readline()
                        if not num_atoms:
                            break
                        num_atoms = int(num_atoms)
                        f.readline()
                        for i in range(num_atoms):
                            f.readline()
                    # actually read the molecule
                    symbols = []
                    positions = []
                    num_atoms = f.readline()
                    if not num_atoms:
                        raise StopIteration
                    num_atoms = int(num_atoms)
                    f.readline()
                    for i in range(num_atoms):
                        line = f.readline()
                        if line.strip():
                            symbol, x, y, z = line.split()[:4]
                            position = (float(x), float(y), float(z))
                            symbols.append(symbol)
                            positions.append(position)
                except StopIteration:
                    raise IndexError("Frame {} not found".format(frame))
            return data.Atoms(positions, None, symbols, self.info.volume)
        except (IOError, IndexError):
            raise
        except Exception as e:
            raise FileError("Cannot read atom data.", e)


class BabelFile(InputFile):
    """
    Implementation on :class:`InputFile` for Open Babel 'xyz' files.
    """

    def __init__(self, path):
        super().__init__(path)
        # Check if the file exists
        f = open(self.path, "r")
        f.close()

    def readinfo(self):
        try:
            file_extension = os.path.splitext(self.path)[1][1:]
            mol_iter = pybel.readfile(file_extension.encode("utf8"), self.path.encode("utf8"))
            try:
                mol = next(mol_iter)
                self._info.volumestr = mol.title
                self._info.num_frames = 1
                for _ in mol_iter:
                    self._info.num_frames += 1
            except StopIteration:
                self._info.num_frames = 0
            self.inforead = True
        except IOError:
            raise
        except Exception as e:
            raise FileError("Cannot read file info.", e)

    def readatoms(self, frame):
        try:
            if self.info.num_frames <= frame:
                raise IndexError("Frame {} not found".format(frame))

            file_extension = os.path.splitext(self.path)[1][1:]
            mol_iter = pybel.readfile(file_extension.encode("utf8"), self.path.encode("utf8"))

            # get the correct frame
            try:
                for _ in range(frame):
                    mol_iter.next()
                mol = mol_iter.next()
            except StopIteration:
                raise IndexError("Frame {} not found".format(frame))

            # read the atom information
            symbols = []
            positions = []
            for atom in mol.atoms:
                positions.append(tuple(float(c) for c in atom.coords))
                symbol = core.elements.symbols[atom.atomicnum]
                symbols.append(symbol)
            return data.Atoms(positions, None, symbols, self.info.volume)
        except (IOError, IndexError):
            raise
        except Exception as e:
            raise FileError("Cannot read atom data.", e)


class ResultFile(InputFile):
    """
    Abstract access to a file that contains both input data (atoms) and
    calculated results.
    The `info` attribute has the type `ResultInfo`.
    In addition the `readinfo` and `readatoms` methods, subclasses must
    implement `writeinfo`, `readresults` and `writeresults`.
    """

    def __init__(self, path, sourcefilepath=None):
        """
        **Parameters:**
            `path` :
                absolute path to the file

            `sourcefilepath` :
                path to the file where the input data originally came from
        """
        super().__init__(path)
        self._info = data.ResultInfo()
        self._info.sourcefilepath = sourcefilepath

    def getresults(self, frame, resolution):
        """
        Read results from this file.

        **Parameters:**
            `frame` :
                the frame number
            `resolution` :
                the resolution of the calculation

        **Returns:**
            A `Results` object, if the file contains results for the
            specified parameters. Otherwise it return `None`.

        **Raises:**
            - :class:`FileError`: if there are problems with the data in the file
            - :class:`IOError`: if the file cannot be read
        """
        if not self.info[resolution].domains[frame] is None:
            return self.readresults(frame, resolution)
        else:
            return None

    def addresults(self, results, overwrite=True):
        """
        Write calculated results into this file.

        **Parameters:**
            `results` :
                the results to write
            `overwrite` :
                specifies if existing results should be overwritten

        **Raises:**
            - :class:`FileError`: if there are problems with the data in the file
            - :class:`IOError`: if the file cannot be read or written
        """
        self.writeresults(results, overwrite=overwrite)
        resinfo = self.info[results.resolution]
        if results.domains:
            resinfo.domains[results.frame] = results.domains.timestamp
        if results.surface_cavities:
            resinfo.surface_cavities[results.frame] = results.surface_cavities.timestamp
        if results.center_cavities:
            resinfo.center_cavities[results.frame] = results.center_cavities.timestamp
        self.writeinfo()

    def writeinfo(self):
        raise NotImplementedError

    def readresults(self, frame, resolution):
        raise NotImplementedError

    def writeresults(self, results, overwrite=True):
        raise NotImplementedError


class HDF5File(ResultFile):
    """
    Implementation on :class:`ResultFile` for 'hdf5' files.
    """

    def __init__(self, path, sourcefilepath=None):
        super().__init__(path, sourcefilepath)

    @classmethod
    def fromInputFile(cls, filepath, sourcefilepath):
        """
        Create a new :class:`HDF5File`. Copy atoms from an
        :class:`InputFile`. This is used to initially create a file
        to export results to.
        """
        inputfile = File.open(sourcefilepath)
        outputfile = cls(filepath, sourcefilepath)
        for frame in range(inputfile.info.num_frames):
            atoms = inputfile.getatoms(frame)
            results = data.Results(filepath, frame, 64, atoms, None, None, None)
            outputfile.writeresults(results)
        outputfile.readinfo()
        outputfile.writeinfo()
        return outputfile

    def readatoms(self, frame):
        atoms = None
        if not os.path.isfile(self.path):
            raise IOError(2, "File not found.")
        try:
            with h5py.File(self.path) as f:
                group = "atoms/frame{}".format(frame)
                if group not in f:
                    raise IndexError("Frame {} not found".format(frame))
                atoms = data.Atoms(f[group])
        except (IOError, IndexError):
            raise
        except Exception as e:
            raise FileError("Cannot read atom data.", e)
        return atoms

    def readinfo(self):
        if not os.path.isfile(self.path):
            raise IOError(2, "File not found.")
        try:
            with h5py.File(self.path) as f:
                if "info" in f:
                    info = data.ResultInfo(f["info"])
                    if self._info.sourcefilepath is not None:
                        info.sourcefilepath = self._info.sourcefilepath
                    self._info = info
                    self.inforead = True
        except IOError:
            raise
        except Exception as e:
            raise FileError("Cannot read file info.", e)

        if not self.inforead and self._info.sourcefilepath is not None and os.path.isfile(self._info.sourcefilepath):
            try:
                sf = File.open(self._info.sourcefilepath)
                self._info.num_frames = sf.info.num_frames
                self._info.volumestr = sf.info.volumestr
                self.inforead = True
            except IOError:
                raise
            except Exception as e:
                raise FileError("Cannot read file info.", e)

        if not self.inforead:
            raise RuntimeError("No File Info in this file and the source file.")

    def writeinfo(self):
        try:
            with h5py.File(self.path, "a") as f:
                h5group = f.require_group("info")
                self.info.tohdf(h5group)
        except IOError:
            raise
        except Exception as e:
            raise FileError("Cannot write file info.", e)

    def readresults(self, frame, resolution):
        if not os.path.isfile(self.path):
            raise IOError(2, "File not found.")
        try:
            results = None
            with h5py.File(self.path) as f:
                groupname = "results/frame{}/resolution{}".format(frame, resolution)
                if groupname in f:
                    group = f[groupname]
                    atoms = data.Atoms(f["atoms/frame{}".format(frame)])
                    domains = data.Domains(group["domains"])
                    if "surface_cavities" in group:
                        surface_cavities = data.Cavities(group["surface_cavities"])
                    else:
                        surface_cavities = None
                    if "center_cavities" in group:
                        center_cavities = data.Cavities(group["center_cavities"])
                    else:
                        center_cavities = None
                    if self.info.sourcefilepath is not None:
                        filepath = self.info.sourcefilepath
                    else:
                        filepath = self.path
                    results = data.Results(
                        filepath,
                        frame,
                        resolution,
                        atoms,
                        domains,
                        surface_cavities,
                        center_cavities,
                    )
        except IOError:
            raise
        except Exception as e:
            raise FileError("Cannot read results.", e)
        return results

    def writeresults(self, results, overwrite=True):
        # results valid? I think so!
        try:
            with h5py.File(self.path, "a") as f:
                group = f.require_group("atoms/frame{}".format(results.frame))
                # is it OK to never overwrite atoms? I think they should be saved when recalculating the file!
                results.atoms.tohdf(group, overwrite=overwrite)
                if (
                    results.domains is not None
                    or results.surface_cavities is not None
                    or results.center_cavities is not None
                ):
                    group = f.require_group("results/frame{}/resolution{}".format(results.frame, results.resolution))
                if results.domains is not None:
                    subgroup = group.require_group("domains")
                    results.domains.tohdf(subgroup, overwrite=overwrite)
                if results.surface_cavities is not None:
                    subgroup = group.require_group("surface_cavities")
                    results.surface_cavities.tohdf(subgroup, overwrite=overwrite)
                if results.center_cavities is not None:
                    subgroup = group.require_group("center_cavities")
                    results.center_cavities.tohdf(subgroup, overwrite=overwrite)
        except IOError:
            raise
        except Exception as e:
            raise FileError("Cannot write results.", e)


class File(object):
    """
    Provides static methods for easy access to files and directories.
    The class attribute `types` associates filename endings with
    classes to handle them.
    """

    types = dict(
        list(zip(pybel.informats.keys(), repeat(BabelFile)))
        + [
            ("xyz", XYZFile),
            ("hdf5", HDF5File),
        ]
    )

    @classmethod
    def listdir(cls, directory):
        """
        List all (possible) pyMolDyn files in the directory.

        **Parameters:**
            `directory` :
                path to a directory

        **Returns:**
            A list of filenames which can be opened.
        """
        if not directory:
            directory = "."
        return [f for f in os.listdir(directory) if cls.exists(f)]

    @classmethod
    def open(cls, filepath):
        """
        Get the associated :class:`InputFile` object for the given file.

        **Parameters:**
            `filepath` :
                path to the file

        **Returns:**
            An object of a subclass of :class:`InputFile`.

        **Raises:**
            :class:`ValueError` if the file format is unknown.
        """
        e = filepath.split(".")[-1]
        if e not in cls.types:
            raise ValueError("Unknown file format")
        FileClass = cls.types[e]
        return FileClass(filepath)

    @classmethod
    def exists(cls, filepath):
        """
        Check if a file exists and if it can be opened.

        **Parameters:**
            `filepath` :
                path to the file

        **Returns:**
            `True` if the file exists and there is a subclass of :class:`InputFile`
            associated with the filename ending.
        """
        filepath = get_abspath(filepath)
        name = os.path.basename(filepath)
        return os.path.isfile(filepath) and name.split(".")[-1] in cls.types
